#!/usr/bin/perl -w
#
#
#
use Fcntl;
use strict;
use POSIX qw(:termios_h);
use IO::Handle;
use IO::Select;
use Getopt::Std;
use vars qw($opt_s $opt_l);
require 'termios.ph';

getopts('l:s:');

$| = 1;	# set AUTOFLUSH

sub get_tty_device()
{
	if (!defined $opt_l) {
		return "/dev/ttyUSB0";
	} elsif (-c $opt_l) {
		return $opt_l;
	} else {
		print "not a character device '$opt_l'\n";
		exit 1;
	}	
}	
sub get_baud_rate()
{
	my $baud = int(((defined $opt_s ? $opt_s : 1200) - 1) / 1200);
	if ($baud >= (115200/1200)) {
		print "setting baudrate to 115200\n";
		return  &POSIX::B115200;
	} elsif ($baud >= (38400/1200)) {
		print "setting baudrate to 38400\n";
		return  &POSIX::B38400;
	} elsif ($baud >= (19200/1200)) {
		print "setting baudrate to 19200\n";
		return  &POSIX::B19200;
	} elsif ($baud >= (9600/1200)) {
		print "setting baudrate to 9600\n";
		return  &POSIX::B9600;
	} elsif ($baud >= (4800/1200)) {
		print "setting baudrate to 4800\n";
		return  &POSIX::B4800;
	} elsif ($baud >= (2400/1200)) {
		print "setting baudrate to 2400\n";
		return  &POSIX::B2400;
	} else {
		print "setting baudrate to 1200\n";
		return  &POSIX::B1200;
	}
}

print "starting ...\n" ;

my $tty = get_tty_device();

print "sysopen ...\n";
sysopen(TTYOUT, $tty, O_RDWR|O_NONBLOCK) || die "can't open TTYOUT $tty - error = $!\n";
print "sysopen ...\n";
sysopen(TTYIN,  $tty, O_RDWR|O_NONBLOCK) || die "can't open TTYIN  $tty - error = $!\n";

print "fileno ...\n";
my $fd_tty_out = fileno TTYOUT;
my $fd_tty_in  = fileno TTYIN;
my $fd_console = fileno STDIN;

print "select ...\n";
my $select = IO::Select->new();

$select->add(\*STDIN);
$select->add(\*TTYIN);

my $term_out = POSIX::Termios->new();
my $term_in  = POSIX::Termios->new();
my $console  = POSIX::Termios->new();

$term_out->getattr($fd_tty_out) || die "failed getattr on fd_tty_out  - error = $!\n";
$term_in->getattr($fd_tty_in) || die "failed getattr on fd_tty_in  - error = $!\n";
$console->getattr($fd_console) || die "failed getattr on fd_console  - error = $!\n";

#
# man POSIX
# man termios
#

my $speed = get_baud_rate();
$term_out->setospeed($speed);
$term_out->setispeed($speed);
{
	my $cterm_out = $term_out->getcflag() || die "failed getcflag on term_out - error = $!\n";
	my $cterm_in = $term_in->getcflag()   || die "failed getcflag  on term_in - error = $!\n";
	my $cset = 0;
	my $czap = 0;
	$cset |= 020000000000;	# CRTSCTS;
	$cset |= &POSIX::CS8;
	$cset |= &POSIX::HUPCL;
	$cset |= &POSIX::CREAD;
	$cterm_out = ~$czap & $cterm_out | $cset;
	$cterm_in  = ~$czap & $cterm_in  | $cset;
	$term_out->setcflag($cterm_out);	# no error code
	$term_in->setcflag($cterm_in);		# no error code
}
{
	my $iterm_out = $term_out->getiflag() || die "failed getiflag on term_out - error = $!\n";
	my $iterm_in = $term_in->getiflag()   || die "failed getiflag  on term_in - error = $!\n";
	my $iset = 0;
	my $izap = 0;
	$izap |= &POSIX::ICRNL;
	$iterm_out = ~$izap & $iterm_out | $iset;
	$iterm_in  = ~$izap & $iterm_in  | $iset;
	$term_out->setiflag($iterm_out);	# no error code
	$term_in->setiflag($iterm_in);		# no error code
}
{
	my $oterm_out = $term_out->getoflag() || die "failed getlflag on term_out - error = $!\n";
	my $oterm_in = $term_in->getoflag()   || die "failed getlflag  on term_in - error = $!\n";
	my $oconsole = $console->getoflag()   || die "failed getlflag  on console - error = $!\n";
	my $oset = 0;
	my $ozap = 0;
	$ozap |= &POSIX::OPOST;
	$oset |= &ONLCR;
	$ozap |= &OLCUC;
	$ozap |= &OCRNL;
	$ozap |= &ONOCR;
	$ozap |= &ONLRET;
	$ozap |= &OFILL;
	$ozap |= &OFDEL;
	$ozap |= &NLDLY;
	$oset |= &NL0;
	$ozap |= &CRDLY;
	$oset |= &CR0;
	$ozap |= &TABDLY;
	$oset |= &TAB0;
	$ozap |= &BSDLY;
	$oset |= &BS0;
	$ozap |= &VTDLY;
	$oset |= &VT0;
	$ozap |= &FFDLY;
	$oset |= &FF0;
	$oterm_out = ~$ozap & $oterm_out | $oset;
	$oterm_in  = ~$ozap & $oterm_in  | $oset;
	$oconsole  = ~$ozap & $oconsole  | $oset;
	$term_out->setoflag($oterm_out);	# no error code
	$term_in->setoflag($oterm_in);		# no error code
	$console->setoflag($oconsole);		# no error code
}
{
	my $lterm_out = $term_out->getlflag() || die "failed getlflag on term_out - error = $!\n";
	my $lterm_in = $term_in->getlflag()   || die "failed getlflag  on term_in - error = $!\n";
	my $lconsole = $console->getlflag()   || die "failed getlflag  on console - error = $!\n";
	my $lset = 0;
	my $lzap = 0;
	$lzap |= &POSIX::ECHO;
	$lzap |= &POSIX::ECHOE;
	$lzap |= &POSIX::ECHONL;
	$lzap |= &POSIX::ECHOK;
	$lzap |= &POSIX::ICANON;
	$lzap |= &POSIX::ISIG;
	$lzap |= &POSIX::NOFLSH;
	$lzap |= &XCASE;
	$lzap |= &POSIX::TOSTOP;
	$lzap |= &ECHOPRT;
	$lset |= &ECHOCTL;
	$lset |= &ECHOKE;
	$lterm_out = ~$lzap & $lterm_out | $lset;
	$lterm_in  = ~$lzap & $lterm_in  | $lset;
	$lconsole  = ~$lzap & $lconsole  | $lset;
	$term_out->setlflag($lterm_out);	# no error code
	$term_in->setlflag($lterm_in);		# no error code
	$console->setlflag($lconsole);		# no error code
}
{
	my $oconsole = $console->getoflag()   || die "failed getlflag  on console - error = $!\n";
	my $oset = 0;
	my $ozap = 0;
	$ozap |= &POSIX::OPOST;
	$oset |= &ONLCR;
	$ozap |= &OLCUC;
	$ozap |= &OCRNL;
	$ozap |= &ONOCR;
	$ozap |= &ONLRET;
	$ozap |= &OFILL;
	$ozap |= &OFDEL;
	$ozap |= &NLDLY;
	$oset |= &NL0;
	$ozap |= &CRDLY;
	$oset |= &CR0;
	$ozap |= &TABDLY;
	$oset |= &TAB0;
	$ozap |= &BSDLY;
	$oset |= &BS0;
	$ozap |= &VTDLY;
	$oset |= &VT0;
	$ozap |= &FFDLY;
	$oset |= &FF0;
	$oconsole  = ~$ozap & $oconsole  | $oset;
	$console->setoflag($oconsole);		# no error code
}
{
	my $lconsole = $console->getlflag()   || die "failed getlflag  on console - error = $!\n";
	my $lset = 0;
	my $lzap = 0;
	$lzap |= &POSIX::ECHO;
	$lzap |= &POSIX::ECHOE;
	$lzap |= &POSIX::ECHONL;
	$lzap |= &POSIX::ECHOK;
	$lzap |= &POSIX::ICANON;
	$lset |= &POSIX::ISIG;
	$lzap |= &POSIX::NOFLSH;
	$lzap |= &XCASE;
	$lzap |= &POSIX::TOSTOP;
	$lzap |= &ECHOPRT;
	$lset |= &ECHOCTL;
	$lset |= &ECHOKE;
	$lconsole  = ~$lzap & $lconsole  | $lset;
	$console->setlflag($lconsole);		# no error code
}
print "setcc ...\n";
$term_out->setcc(&POSIX::VMIN,1);	# no error code
$term_in->setcc(&POSIX::VMIN,1);	# no error code
$console->setcc(&POSIX::VMIN,1);	# no error code

$term_out->setcc(&POSIX::VTIME,0);	# no error code
$term_in->setcc(&POSIX::VTIME,0);	# no error code
$console->setcc(&POSIX::VTIME,0);	# no error code

print "setattr ...\n";
$term_out->setattr($fd_tty_out,&POSIX::TCSANOW)	|| die "failed setattr on term_out - error = $!\n";
$term_in->setattr($fd_tty_in,&POSIX::TCSANOW)	|| die "failed setattr  on term_in - error = $!\n";
$console->setattr($fd_console,&POSIX::TCSANOW)	|| die "failed setattr  on console - error = $!\n";

my $message = "hello there how are you\n";
syswrite(TTYOUT,$message);

while(1) {
	my @ready = $select->can_read(1.0);
	if (0 == scalar @ready) {
	} elsif ($ready[0] == \*STDIN) {
		my $buffer;
		my $read = sysread STDIN,$buffer,128 || die "failed sysread STDIN - error = $!\n";
		my $write = syswrite TTYOUT,$buffer;
	} elsif ($ready[0] == \*TTYIN) {
		my $buffer;
		my $read = sysread TTYIN,$buffer,128 || die "failed sysread TTYIN - error = $!\n";
		print "$buffer";
	} else {
	}
}
exit 0;
