/*
 * Author: Tony Olech - Elan Digital Systems Limited
 * $Revision: 1.36 $
 * $Date: 2007/12/20 15:06:18 $
 * 
 * VMB register declarations and HCD data structures
 *
 * Copyright (C) 2005 Elan Digital Systems Limited
 */

#define VMB_FIFO_SHIFT 14
#define VMB_FIFO_SIZE (1<<VMB_FIFO_SHIFT)
#define VMB_FIFO_MASK ((1<<VMB_FIFO_SHIFT)-1)

#define VMB_URB_STATE_RETURNED_TO_CALLER  0
#define VMB_URB_STATE_QUEUED_FOR_ACTION   1
#define VMB_URB_STATE_CONTROL_SETUP      70
#define VMB_URB_STATE_CONTROL_OUT        71
#define VMB_URB_STATE_CONTROL_IN         72
#define VMB_URB_STATE_CONTROL_ACK_OUT    73
#define VMB_URB_STATE_CONTROL_ACK_IN     74
#define VMB_URB_STATE_CONTROL_STALL      75
#define VMB_URB_STATE_CONTROL_ERR        76
#define VMB_URB_STATE_CONTROL_IN_FAILED  77
#define VMB_URB_STATE_BULK_OUT           62
#define VMB_URB_STATE_BULK_IN            63
#define VMB_URB_STATE_BULK_IN_STALL      64
#define VMB_URB_STATE_BULK_OVERSPILL     65
#define VMB_URB_STATE_BULK_IN_FAILED     66
#define VMB_URB_STATE_ISO_OUT            96
#define VMB_URB_STATE_ISO_IN             97
#define VMB_URB_STATE_INTERRUPT          86
#define VMB_URB_STATE_INT_WAIT           85
#define VMB_URB_STATE_INT_FAILED         84

enum vmb_port_status {

	VPS_error,				// 0
	VPS_monitoring,			// 1
	VPS_detected,			// 2
	VPS_resetting,			// 3
	VPS_enabled,			// 4
	VPS_enumerating,		// 5
	VPS_running,			// 6
};
#define MAX_BITS_PER_FRAME 11960	// 0x2EB8
#define PACKET_BITS_SETUP_IN(bytes) ((32+3+4) + (32+(bytes)*8+3+4) + (16+3+4))
#define PACKET_BITS_SETUP_OUT(bytes) ((32+3+4) + (32+(bytes)*8+3+4) + (16+3+4))
#define PACKET_BITS_DATA_IN(bytes) ((32+3+4) + (32+(bytes)*8*7/6+3+9) + (16+3+4))
#define PACKET_BITS_DATA_OUT(bytes) ((32+3+4) + (32+(bytes)*8*7/6+3+4) + (16+3+4))
#define UTL_FRAME_SHIFT 11
#define UTL_FRAME_SIZE (1<<UTL_FRAME_SHIFT)
#define UTL_FRAME_MASK ((1<<UTL_FRAME_SHIFT)-1)
struct vmb_frame {

	//unsigned written:1;
	//unsigned padding:15;

	u16 txfifo_bytes;	// diagnostics at the moment
	unsigned fifo_offset:VMB_FIFO_SHIFT;
	u16 scheduled_bits;	// total of all writes in this frame --- max 11960 ==> width = 14
	u16 requested_bits; // total of all reads in this frame
	//u16 input_space;
};
#define MAX_VMB_PORTS	2
struct vmb_port {

	struct usb_device*udev;	// that is attached
	struct vmb *vmb;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
	struct work_struct status_work;
#else
	struct delayed_work delayed_work;
#endif	
	enum vmb_port_status status;
	u8	number;
	u32	Status;
	unsigned	enable:1;	// reset finished
	unsigned	device:1;	// usb device detected on this port (whilst SR.Enable clear)
	unsigned	reset:1;	// port being reset
	unsigned	power:1;
	unsigned long finish;
};
struct vmb_event {

	struct vmb *vmb;
	struct work_struct work;
	u32 kicked;
	u32 worked;
};
//#define SLOT_FRAME_SHIFT 5
//#define SLOT_FRAME_SIZE (1<<SLOT_FRAME_SHIFT)
//#define SLOT_FRAME_MASK ((1<<SLOT_FRAME_SHIFT)-1)
//struct slot_frame {
//
//	u16 erst_id;
//	u16 last_id;
//};
#define SLOT_IDREF_SHIFT 2
#define SLOT_IDREF_SIZE (1<<SLOT_IDREF_SHIFT)
#define SLOT_IDREF_MASK ((1<<SLOT_IDREF_SHIFT)-1)
struct slot_idref {

	u8  queue_this;	// unsigned queue_this:6
	u8  spare_bits;	// unsigned spare_bits:4
	u16 frame_base;	// unsigned frame_base:11
	u16 frame_last;	// unsigned frame_last:11
};
#define SLOT_QUEUE_SHIFT (SLOT_IDREF_SHIFT+4)
#define SLOT_QUEUE_SIZE (1<<SLOT_QUEUE_SHIFT)
#define SLOT_QUEUE_MASK ((1<<SLOT_QUEUE_SHIFT)-1)
#define	MAX_VMB_SLOTS	16
//
// actioned flag is used for the situation where a CONTROL_SETUP
//               packet has been put in the outgoing fifo for a
//               particular frame, so that when scheduling later
//               frames the CONTROL_SETUP is not retransmitted.
//               This applies also to CONTROL_ACK, INTERRUPT OUT,
//               but not to BULK and ISOCHRONOUS OUT's nor any
//               type of IN
//
//
struct vmb_slot {

	u8 td;
	u8 retries;
	u16 recieved;
	u16 write__id;
	unsigned pipetype:2;
	unsigned pipeout:1;
	unsigned allocated:1;	// set possibly in INTERRUPT context in thread from API
	unsigned initialized:1;	// set by scheduler when holding SW_LOCK
	unsigned retired:1;
	unsigned running:1;		// set when the TLER is set active and cleared when known to be not active
	unsigned actioned:1;	// set when data for queued urb has been sent (or requested ??)
	unsigned end_seen:1;	// an OUT or SETUP has retired via EOT or TLER command
	unsigned retry_flag:1;
	unsigned device_error:1;
	unsigned ec:3;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
	struct work_struct status_work;
#else
	struct delayed_work delayed_work;
#endif
	int line_number; // diagnostic for work queue caller
	unsigned long  jiffies;
	struct vmb  *vmb;
	u8 next_urb_index;	// next urb to be used
	u16 next_urb_count;	// offset within --"--
	u8 idref_done;	// incremented after finished with
	u16 idref_this;	// the current idref or the next one just after finishing it
	u16 idref_pkts;	// number of packets written to fifo from urb referenced by idref_this
	struct slot_idref idref[SLOT_IDREF_SIZE];
	u8 queue_size;	// number of urb's
	u16 queue_next; // vacant urb index
	struct urb  *u_r_b[SLOT_QUEUE_SIZE];
	u8 toggle[SLOT_QUEUE_SIZE];
//	u16 last_frame;	// being actioned by the VMB hardware
//	u16 next_frame;	// to be scheduled by this driver
//	struct slot_frame frame[SLOT_FRAME_SIZE];	// tied to vmb_slot
	u8 usb_addr;
	u8 usb_endp;
	u16 max_plen;
	u16 maxusbpktlen;
	u16 extra;
	u8 overspill[2048];
};
struct vmb_udev {

	struct usb_device*usb_device;	// that is attached
	// the port used by devices that are connected via hubs 
	// cannot be determined - only direct connection can
	u8	portnumber;	// if known otherwise 0
	u8	devicefail;
};
#define MAX_DEV_ADDRS 128

#define URB_GIVEBACK_SHIFT (SLOT_QUEUE_SHIFT+1)
#define URB_GIVEBACK_SIZE (1<<URB_GIVEBACK_SHIFT)
#define URB_GIVEBACK_MASK ((1<<URB_GIVEBACK_SHIFT)-1)

#define URB_DE_QUEUE_SHIFT (SLOT_QUEUE_SHIFT+1)
#define URB_DE_QUEUE_SIZE (1<<URB_DE_QUEUE_SHIFT)
#define URB_DE_QUEUE_MASK ((1<<URB_DE_QUEUE_SHIFT)-1)

#define URB_EN_QUEUE_SHIFT (SLOT_QUEUE_SHIFT+2)
#define URB_EN_QUEUE_SIZE (1<<URB_EN_QUEUE_SHIFT)
#define URB_EN_QUEUE_MASK ((1<<URB_EN_QUEUE_SHIFT)-1)
struct vmb {
	struct list_head	vmb_list;
	spinlock_t		 hw_lock;
	struct semaphore sw_lock;
#if LOCK_DEBUGGING_COMPILED_IN
	int   hw_flag;	// for ASSERTION's
	int   sw_flag;	// for ASSERTION's
	int   hw_line;	// for ASSERTION's
	int   sw_line;	// for ASSERTION's
#endif	
	void __iomem		*addr;	// mem
	void __iomem 		*base;	// iop
	struct vmb_platform_data	  *board;
	struct proc_dir_entry   *pde;
	unsigned	power:1;
	unsigned	reset:1;
	unsigned	unplugged:1;
	unsigned	enumerating:1;
	u16			 __frame__;
	u16			diag_txit;	// for diagnostics
	u16			diag_txwp;	// for diagnostics
	u16			diag_txrp;	// for diagnostics
	u16			diag_rxwp;	// for diagnostics
	u16			diag_rxrp;	// for diagnostics
	unsigned long	   stat_lost;
	u8			current_port;	// for enumeration
	u32			next_slot;
	struct usb_device *current_udev; // for enumeration
	int			slot_count;	// 0 -- 16

	u8	sr_mask;	// protected by hw_lock

	u16 frame_tail;
	u16 frame_head;

	struct list_head	async;
	struct work_struct	scheduler;
#if DEBUGGING_COMPILED_IN
	int last_scheduled_from_line;
#define KICK_SCHEDULER(vmb)	do{ \
	vmb->last_scheduled_from_line = __LINE__; \
	queue_work(workqueue,&vmb->scheduler); \
}while(0)
#else
#define KICK_SCHEDULER(vmb)	queue_work(workqueue,&vmb->scheduler)
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
	struct work_struct diagnostics;
#else
	struct delayed_work diagnostics;
#endif
	struct work_struct	urb_en_queue;
	struct work_struct	urb_de_queue;

	struct vmb_event rxthresh,rxto,txthresh,ovr,eot,rxerr;

	u16 en_queue_size;
	u16 en_queue_next;

	u16 de_queue_size;
	u16 de_queue_next;

	u16 giveback_size;
	u16 giveback_next;

	struct urb*en_queue[URB_EN_QUEUE_SIZE];
	struct urb*de_queue[URB_DE_QUEUE_SIZE];
	struct urb*giveback[URB_GIVEBACK_SIZE];
	struct vmb_udev udev[MAX_DEV_ADDRS];
	struct vmb_port	port[MAX_VMB_PORTS];
	struct vmb_slot slot[MAX_VMB_SLOTS];
	struct vmb_frame frame[UTL_FRAME_SIZE];
	
};
inline char*feature_string(int feature)
{
	switch(feature)
	{
	case USB_PORT_FEAT_CONNECTION: return "CONNECTION";
	case USB_PORT_FEAT_ENABLE: return "ENABLE";
	case USB_PORT_FEAT_SUSPEND: return "SUSPEND";
	case USB_PORT_FEAT_OVER_CURRENT: return "OVER_CURRENT";
	case USB_PORT_FEAT_RESET: return "RESET";
	case USB_PORT_FEAT_POWER: return "POWER";
	case USB_PORT_FEAT_LOWSPEED: return "LOWSPEED";
	case USB_PORT_FEAT_HIGHSPEED: return "HIGHSPEED";
	case USB_PORT_FEAT_C_CONNECTION: return "C_CONNECTION";
	case USB_PORT_FEAT_C_ENABLE: return "C_ENABLE";
	case USB_PORT_FEAT_C_SUSPEND: return "C_SUSPEND";
	case USB_PORT_FEAT_C_OVER_CURRENT: return "C_OVER_CURRENT";
	case USB_PORT_FEAT_C_RESET: return "C_RESET";
	case USB_PORT_FEAT_TEST: return "TEST";
	case USB_PORT_FEAT_INDICATOR: return "INDICATOR";
	default: return "****unknown****";
	}
}

static void slot_assertion_failure(const char*function,const char*filename,const unsigned line,const char*failure,struct vmb_slot *slot);
static inline struct vmb *hcd_to_vmb(struct usb_hcd *hcd)
{
	return (struct vmb *) (hcd->hcd_priv);
}
static inline struct vmb *urb_to_vmb(struct urb *urb)
{
	ASSERT(urb);

	struct vmb	*vmb;

	list_for_each_entry(vmb,&vmb_static_list,vmb_list) {

		if (vmb == urb->hcpriv) {

			return vmb;
		}
	}
	ASSERT(0);

	return (struct vmb *) urb->hcpriv;
}
static inline int check_urb_is_in_slot(struct vmb_slot *slot, struct urb *urb)
{
	ASSERT(urb);
	SLOT_ASSERT(slot);
	ASSERT(slot->vmb == urb->hcpriv);
	SLOT_ASSERT(slot->allocated);

	u8 queue_size = slot->queue_size;
	u16 queue_next = slot->queue_next;

	while(queue_size-- > 0) {

		if (urb == slot->u_r_b[SLOT_QUEUE_MASK & --queue_next]) {
			return 1;
		} else {
			continue;
		};
	}
	return 0;
}
static inline struct vmb_slot *urb_to_slot(struct urb *urb)
{
	ASSERT(urb);

	struct vmb	*vmb;

	list_for_each_entry(vmb,&vmb_static_list,vmb_list) {

		if (vmb == urb->hcpriv) {

			int slots = MAX_VMB_SLOTS;

			while(slots-- > 0)
			{
				struct vmb_slot *slot = &vmb->slot[slots];

				ASSERT(slot->vmb == vmb);
				ASSERT(slot->td == slots);

				if (slot->initialized || slot->allocated) {
					SLOT_ASSERT(slot->allocated);
					u8 queue_size = slot->queue_size;
					u16 queue_next = slot->queue_next;

					while(queue_size-- > 0) {

						if (urb == slot->u_r_b[SLOT_QUEUE_MASK & --queue_next]) {
							return slot;
						} else {
							continue;
						};
					}
					continue;
				} else {
					continue;
				}
			}
		}
	}
	ASSERT(0);

	return (struct vmb_slot *) NULL;
}
static inline int find_urb_position_in_slot(struct vmb_slot *slot, struct urb *urb)
{
	SLOT_ASSERT(urb);
	SLOT_ASSERT(slot->allocated);
	SLOT_ASSERT(slot->initialized);

	u8 queue_size = slot->queue_size;
	u16 queue_next = slot->queue_next;

	do {
		if (urb == slot->u_r_b[SLOT_QUEUE_MASK & --queue_next])
			return queue_size;
	} while (--queue_size > 0);

	return queue_size;
}
static inline struct urb *slot_to_head_urb(struct vmb_slot *slot)
{
	SLOT_ASSERT(slot->allocated);	// might not yet be initialized

	u8 queue_size = slot->queue_size;

	if (queue_size > 0) {

		u16 queue_head = slot->queue_next - queue_size;

		return slot->u_r_b[SLOT_QUEUE_MASK & queue_head];

	} else {

		return (struct urb *) NULL;
	}
}
static inline int remove_slot_head_urb(struct vmb_slot *slot, struct urb *urb)
{
	SLOT_ASSERT(slot->allocated);	// might not yet be initialized

	u8 queue_size = slot->queue_size--;

	SLOT_ASSERT(queue_size > 0);

	u16 queue_head = slot->queue_next - queue_size;

	SLOT_ASSERT(urb == slot->u_r_b[SLOT_QUEUE_MASK & queue_head]);

	slot->u_r_b[SLOT_QUEUE_MASK & queue_head] = NULL;

	return (1 == queue_size);
}
static inline struct vmb *slot_to_vmb(struct vmb_slot *slot)	// TODO == replace this by vmb_slot_to_vmb()
{
	struct vmb	*vmb;

	list_for_each_entry(vmb,&vmb_static_list,vmb_list) {

		int slots = MAX_VMB_SLOTS;

		while(slots-- > 0)
		{
			if (&vmb->slot[slots] == slot) {

				ASSERT(slot->vmb == vmb);
				ASSERT(slot->td == slots);

				return vmb;
			}
		}
	}
	SLOT_ASSERT(0);

	return (struct vmb *) NULL;
}
static inline struct vmb *port_to_vmb(struct vmb_port *port)
{
	struct vmb	*vmb;

	list_for_each_entry(vmb,&vmb_static_list,vmb_list) {

		int ports = MAX_VMB_PORTS;

		while(ports-- > 0)
		{
			if (&vmb->port[ports] == port) {

				ASSERT(port->vmb == vmb);
				ASSERT(port->number == 1+ports);

				return vmb;
			}
		}
	}
	ASSERT(0);

	return (struct vmb *) NULL;
}
static inline struct vmb *event_to_vmb(struct vmb_event *event)
{
	struct vmb	*vmb;

	list_for_each_entry(vmb,&vmb_static_list,vmb_list) {

		if (event->vmb == vmb) {
	
			if (&vmb->rxthresh == event) {	
				return vmb;
			} else if (&vmb->rxto == event) {	
				return vmb;
			} else if (&vmb->txthresh == event) {	
				return vmb;
			} else if (&vmb->ovr == event) {	
				return vmb;
			} else if (&vmb->eot == event) {	
				return vmb;
			} else if (&vmb->rxerr == event) {	
				return vmb;
			} else {
				ASSERT(0);
			}

		} else {

			if (&vmb->rxthresh == event) {	
				ASSERT(0);
			} else if (&vmb->rxto == event) {	
				ASSERT(0);
			} else if (&vmb->txthresh == event) {	
				ASSERT(0);
			} else if (&vmb->ovr == event) {	
				ASSERT(0);
			} else if (&vmb->eot == event) {	
				ASSERT(0);
			} else if (&vmb->rxerr == event) {	
				ASSERT(0);
			} else {
				continue;
			}
		}
	}
	ASSERT(0);

	return (struct vmb *) NULL;
}
static inline int vmb_is_still_valid(struct vmb *tst)
{
	ASSERT(tst);
	struct vmb *vmb;
	list_for_each_entry(vmb,&vmb_static_list,vmb_list) {

		if (vmb == tst) {
			return 1;
		}
	}
    return 0;
}
static inline void assert_valid_vmb(struct vmb *vmb,const char*function)
{
	if (vmb_is_still_valid(vmb)) {	
	} else {
		DBG(0,"%s(vmb=%p is not valid)\n",function,vmb);	
    	ASSERT(0);
	}
}
static inline struct vmb *vmb_slot_to_vmb(struct vmb_slot *slot)
{
	ASSERT(slot);
	u8 td = slot->td;
	ASSERT(td < MAX_VMB_SLOTS);
	struct vmb *vmb = container_of((void *) slot, struct vmb, slot[td]);

	assert_valid_vmb(vmb,__FUNCTION__);

	return vmb;
}
static inline struct vmb *urb_de_queue_to_vmb(struct work_struct*work_struct)
{
	struct vmb *vmb = container_of((void *) work_struct, struct vmb, urb_de_queue);

	assert_valid_vmb(vmb,__FUNCTION__);

	ASSERT(! in_atomic());

	return vmb;
}
static inline struct vmb *urb_en_queue_to_vmb(struct work_struct*work_struct)
{
	struct vmb *vmb = container_of((void *) work_struct, struct vmb, urb_en_queue);

	assert_valid_vmb(vmb,__FUNCTION__);

	ASSERT(! in_atomic());

	return vmb;
}
static inline struct vmb *scheduler_to_vmb(struct work_struct*work_struct)
{
	struct vmb *vmb = container_of((void *) work_struct, struct vmb, scheduler);

	ASSERT(! in_atomic());

	return vmb;
}
static inline struct vmb *rxthresh_to_vmb(struct work_struct*work_struct)
{
	struct vmb *vmb = container_of((void *) work_struct, struct vmb, rxthresh.work);

	assert_valid_vmb(vmb,__FUNCTION__);

	ASSERT(! in_atomic());

	return vmb;
}
static inline struct vmb *rxto_to_vmb(struct work_struct*work_struct)
{
	struct vmb *vmb = container_of((void *) work_struct, struct vmb, rxto.work);

	assert_valid_vmb(vmb,__FUNCTION__);

	ASSERT(! in_atomic());

	return vmb;
}
static inline struct vmb *txthresh_to_vmb(struct work_struct*work_struct)
{
	struct vmb *vmb = container_of((void *) work_struct, struct vmb, txthresh.work);

	assert_valid_vmb(vmb,__FUNCTION__);

	ASSERT(! in_atomic());

	return vmb;
}
static inline struct vmb *ovr_to_vmb(struct work_struct*work_struct)
{
	struct vmb *vmb = container_of((void *) work_struct, struct vmb, ovr.work);

	assert_valid_vmb(vmb,__FUNCTION__);

	ASSERT(! in_atomic());

	return vmb;
}
static inline struct vmb *eot_to_vmb(struct work_struct*work_struct)
{
	struct vmb *vmb = container_of((void *) work_struct, struct vmb, eot.work);

	assert_valid_vmb(vmb,__FUNCTION__);

	ASSERT(! in_atomic());

	return vmb;
}
static inline struct vmb *rxerr_to_vmb(struct work_struct*work_struct)
{
	struct vmb *vmb = container_of((void *) work_struct, struct vmb, rxerr.work);

	assert_valid_vmb(vmb,__FUNCTION__);

	ASSERT(! in_atomic());

	return vmb;
}
static inline struct usb_hcd *vmb_to_hcd(struct vmb *vmb)
{
	return container_of((void *) vmb, struct usb_hcd, hcd_priv);
}
static inline void free_vmb_slot(struct vmb *vmb, struct vmb_slot *slot)
{
	ASSERT(0 < in_vmb_scheduler);	

	ASSERT(vmb == slot->vmb);

	SLOT_ASSERT(0 == slot->idref_done);
	SLOT_ASSERT(0 == slot->retry_flag);
	SLOT_ASSERT(1 == slot->retired);
	SLOT_ASSERT(1 == slot->actioned);
	SLOT_ASSERT(1 == slot->allocated);	// might not yet be initialized
	SLOT_ASSERT(0 == slot->device_error);
	SLOT_ASSERT(0 == slot->queue_size);
	SLOT_ASSERT(0 == slot->usb_addr);
	SLOT_ASSERT(0 == slot->running);
	SLOT_ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
	SLOT_ASSERT(0 == slot->next_urb_count);

	slot->jiffies = jiffies;

	slot->allocated = 0;
	slot->initialized = 0;
	slot->actioned = 0;
	slot->end_seen = 0;
	slot->retry_flag = 0;

	vmb->slot_count += 1;

	return;
}
static inline struct vmb_slot *find_empty_vmb_slot(struct vmb *vmb,struct urb *urb)
{
	ASSERT(0 < in_vmb_scheduler);	

	int slots = MAX_VMB_SLOTS;

	while(slots-- > 0)
	{
		u8 td = 0xF & vmb->next_slot++;
		struct vmb_slot *slot = &vmb->slot[td];

		ASSERT(vmb == slot->vmb);
		ASSERT(td == slot->td);

		if (slot->allocated) {
			continue;
		} else {

			ASSERT(0 < vmb->slot_count);

			vmb->slot_count -= 1;

			ASSERT(0 == slot->initialized);
			ASSERT(0 == slot->actioned);
			ASSERT(0 == slot->idref_done);
			ASSERT(1 == slot->retired);
			ASSERT(0 == slot->usb_addr);
			ASSERT(0 == slot->running);
			ASSERT(0 == slot->retry_flag);
			ASSERT(0 == slot->device_error);

			ASSERT((SLOT_QUEUE_MASK & slot->next_urb_index) == (SLOT_QUEUE_MASK & slot->queue_next));
			ASSERT(0 == slot->next_urb_count);

			slot->end_seen = 0;
			slot->retries = 0;
			slot->recieved = 0;
			slot->retired = 0;
			slot->extra = 0;
			slot->usb_addr = usb_pipedevice(urb->pipe);
			slot->usb_endp = usb_pipeendpoint(urb->pipe);
			slot->pipetype = usb_pipetype(urb->pipe);
			slot->pipeout = usb_pipeout(urb->pipe);
			slot->maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));

			ASSERT(0 == slot->queue_size);

			slot->allocated = 1;	

			return slot;
		}
	}
	return NULL;
}
static inline struct vmb_slot *find_matching_vmb_slot(struct vmb *vmb,u8 usb_addr,u8 usb_endp,u8 pipeout)
{
	int slots = MAX_VMB_SLOTS;
	int n = 0;

	while(slots-- > 0)
	{
		u8 td = n++;
		struct vmb_slot *slot = &vmb->slot[td];

		ASSERT(vmb == slot->vmb);
		ASSERT(td == slot->td);

		if (1 == slot->allocated && usb_addr == slot->usb_addr && usb_endp == slot->usb_endp) {
			if (0 == usb_addr || pipeout == slot->pipeout) {
				return slot;
			} else {
				continue;
			}
		} else {
			continue;
		}
	}
	return NULL;
}
/*-------------------------------------------------------------------------*/
// ignored --> watched --> reset --> enable --> enumerate --> running
enum vmb_device_status
{
	VmbDS_enumerating,
	VDS_unused,
	VDS_p1_ignored_p2_ignored,
	VDS_p1_watched_p2_watched,

	VDS_p1_reset_p2_ignored,
	VDS_p1_enable_p2_ignored,
	VDS_p1_enumerate_p2_ignored,
	VDS_p1_running_p2_watched,

	VDS_p1_ignored_p2_reset,
	VDS_p1_ignored_p2_enable,
	VDS_p1_ignored_p2_enumerate,
	VDS_p1_watched_p2_running,

	VDS_p1_running_p2_reset,
	VDS_p1_running_p2_enable,
	VDS_p1_running_p2_enumerate,

	VDS_p1_reset_p2_running,
	VDS_p1_enable_p2_running,
	VDS_p1_enumerate_p2_running,

	VDS_p1_running_p2_running,
};
/*-------------------------------------------------------------------------*/
enum vmb_urb_status
{
	VUS_unused,
	VUS_setup_write,
};
/*-------------------------------------------------------------------------*/
//			  UTILITY FUNCTIONS
/*-------------------------------------------------------------------------*/
inline u16 set_slot_tle(struct vmb_slot*slot, unsigned t,unsigned a)
{
	ASSERT(slot->allocated);
	ASSERT(slot->initialized);

	u16 tler	= (0x7 & slot->ec) << 13
				| (0xF & slot->usb_endp) << 9
				| (0x7 & slot->usb_addr) << 6
				| (0x7 & t) << 3
				| (0x1 & a)
				;

	return tler;
}
inline u8 vmb_maxpacket_mpl(u16 maxusbpktlen)
{
	u8 mpl = maxusbpktlen > 63 ? 0x3 : maxusbpktlen > 31 ? 0x2 : maxusbpktlen > 15 ? 0x1 : maxusbpktlen > 7 ? 0x0 : 0x3 ;

	return mpl;
}
inline u8 vmb_maxpacket_len(u16 maxusbpktlen)
{
	u8 len = maxusbpktlen > 63 ? 64 : maxusbpktlen > 31 ? 32 : maxusbpktlen > 15 ? 16 : maxusbpktlen > 7 ? 8 : 64 ;

	return len;
}
inline u8 vmb_maxpacket_shift(u16 maxusbpktlen)
{
	u8 shift = maxusbpktlen > 63 ? 6 : maxusbpktlen > 31 ? 5 : maxusbpktlen > 15 ? 4 : maxusbpktlen > 7 ? 3 : 6 ;

	return shift;
}
inline u32 vmb_maxpacket_mask(u16 maxusbpktlen)
{
	u32 mask = maxusbpktlen > 63 ? 0x0000003F : maxusbpktlen > 31 ? 0x0000001F : maxusbpktlen > 15 ? 0x0000000F : maxusbpktlen > 7 ? 0x00000007 : 0x0000003F ;

	return mask;
}
/*-------------------------------------------------------------------------*/
//			  IOPORT MAPPED
/*-------------------------------------------------------------------------*/
inline u8 read_cr(struct vmb*vmb)
{
	ASSERT(vmb->base > 0);	
	u8 cr = inb(0+(int)vmb->base);

	return cr;
}
inline void write_cr(struct vmb*vmb,u8 cr)
{
	ASSERT(vmb->base > 0);	
	outb(cr,0+(int)vmb->base);
}
inline u8 read_sr(struct vmb*vmb)
{
	ASSERT(vmb->base > 0);	
	u8 sr = inb(1+(int)vmb->base);

	return sr;
}
inline void write_sr(struct vmb*vmb,u8 sr)
{
	ASSERT(vmb->base > 0);	
	outb(sr,1+(int)vmb->base);
}
inline u8 read_tl(struct vmb*vmb)
{
	ASSERT(vmb->base > 0);	
	u8 cr = inb(2+(int)vmb->base);

	return cr;
}
inline void write_tl(struct vmb*vmb,u8 cr)
{
	ASSERT(vmb->base > 0);	
	outb(cr,2+(int)vmb->base);
}
inline u8 read_txpc(struct vmb*vmb)
{
	ASSERT(vmb->base > 0);	
	u8 tp = inb(3+(int)vmb->base);

	return tp;
}
inline void write_txpc(struct vmb*vmb,u8 tp)
{
	ASSERT(vmb->base > 0);	
	outb(tp,3+(int)vmb->base);
}
inline u8 read_pcsr(struct vmb*vmb)
{
	ASSERT(vmb->base > 0);	
	u8 pcsr = inb(4+(int)vmb->base);

	return pcsr;
}
inline void write_pcsr(struct vmb*vmb,u8 pcsr)
{
	ASSERT(vmb->base > 0);	
	outb(pcsr,4+(int)vmb->base);
}
inline u16 read_fc(struct vmb*vmb)
{
	ASSERT(vmb->base > 0);	
	HW_ASSERT_LOCKED(vmb);

	u8 fclr = inb(5+(int)vmb->base);
	u8 fchr = inb(6+(int)vmb->base) & 0x07;
	u8 fchk = inb(5+(int)vmb->base);

	if (fchk == fclr) {
		return fclr | (fchr << 8);
	} else {
		u8 frpt = inb(6+(int)vmb->base) & 0x07;
		return fchk | (frpt << 8);
	}
}
/*-------------------------------------------------------------------------*/
//			  MEMORY MAPPED
/*-------------------------------------------------------------------------*/
inline u8 read_core(struct vmb*vmb)
{
	u8 core = readb((void*)(0x200 + vmb->addr));

	return core;
}
inline void write_core(struct vmb*vmb,u8 core)
{
	writeb(core,(void*)(0x200 + vmb->addr));
}
inline u8 read_fifo(struct vmb*vmb)
{
	u8 fifo = readb((void*)(0x240 + vmb->addr));

	return fifo;
}
inline void write_fifo(struct vmb*vmb,u8 fifo)
{
	writeb(fifo,(void*)(0x240 + vmb->addr));
}
#define READ_TLER(v,n) vmb_read_tler(__FUNCTION__,__LINE__,v,n)
inline void write_tler(struct vmb*vmb,u8 n,u16 tler)	// 0 =< n < 16
{
	ASSERT(n<16);

	writew(tler,(void*)(0x0360+n+n+vmb->addr));
}
inline u8 read_rxrplr(struct vmb*vmb)
{
	u8 rxrplr = readb((void*)(0x0342 + vmb->addr));

	return rxrplr;
}
inline u8 read_rxrphr(struct vmb*vmb)
{
	u8 rxrphr = readb((void*)(0x0344 + vmb->addr));

	return rxrphr;
}
inline u16 read_rxrp(struct vmb*vmb)
{
	u8 rxrplr = readb((void*)(0x0342 + vmb->addr));
	u8 rxrphr = 0x3F & readb((void*)(0x0344 + vmb->addr));

	return rxrplr | (rxrphr << 8);
}
inline u8 read_rxwplr(struct vmb*vmb)
{
	u8 rxwplr = readb((void*)(0x0346 + vmb->addr));

	return rxwplr;
}
inline u8 read_rxwphr(struct vmb*vmb)
{
	u8 rxwphr = readb((void*)(0x0348 + vmb->addr));

	return rxwphr;
}
inline u16 read_rxwp(struct vmb*vmb)
{
	u8 rxwplr = readb((void*)(0x0346 + vmb->addr));
	u8 rxwphr = 0x3F & readb((void*)(0x0348 + vmb->addr));

	return rxwplr | (rxwphr << 8);
}
//inline u8 read_txrplr(struct vmb*vmb)
//{
//	u8 txrplr = readb((void*)(0x034A + vmb->addr));
//
//	return txrplr;
//}
//inline u8 read_txrphr(struct vmb*vmb)
//{
//	u8 txrphr = readb((void*)(0x034C + vmb->addr));
//
//	return txrphr;
//}
inline u16 read_txrp(struct vmb*vmb)
{
	u8 txrplr = readb((void*)(0x034A + vmb->addr));	// LOW byte must be first
	u8 txrphr = 0x3F & readb((void*)(0x034C + vmb->addr));

	return ((u16)txrphr << 8) | (u16)txrplr ;
}
//inline u8 read_txwplr(struct vmb*vmb)
//{
//	u8 txwplr = readb((void*)(0x034E + vmb->addr));
//
//	return txwplr;
//}
//inline u8 read_txwphr(struct vmb*vmb)
//{
//	u8 txwphr = readb((void*)(0x0350 + vmb->addr));
//
//	return txwphr;
//}
inline u16 read_txwp(struct vmb*vmb)
{
	u8 txwplr = readb((void*)(0x034E + vmb->addr));
	u8 txwphr = 0x3F & readb((void*)(0x0350 + vmb->addr));

	return ((u16)txwphr << 8) | (u16)txwplr ;
}
inline u16 fifo_space(struct vmb*vmb)
{
	u16 txrp = read_txrp(vmb);
	u16 txwp = read_txwp(vmb);

	return txrp > txwp ? txrp - txwp : VMB_FIFO_SIZE + txrp - txwp ;
}
inline u8 read_rxitlr(struct vmb*vmb)
{
	u8 rxitlr = readb((void*)(0x0352 + vmb->addr));

	return rxitlr;
}
inline void write_rxitlr(struct vmb*vmb,u8 rxitlr)
{
	writeb(rxitlr,(void*)(0x0352 + vmb->addr));
}
inline u8 read_rxithr(struct vmb*vmb)
{
	u8 rxithr = readb((void*)(0x0354 + vmb->addr));

	return rxithr;
}
inline void write_rxithr(struct vmb*vmb,u8 rxithr)
{
	writeb(rxithr,(void*)(0x0354 + vmb->addr));
}
inline u16 read_rxit(struct vmb*vmb)
{
	u8 rxitlr = readb((void*)(0x0352 + vmb->addr));
	u8 rxithr = 0x3F & readb((void*)(0x0354 + vmb->addr));

	return rxitlr | (rxithr << 8);
}
inline u8 read_txitlr(struct vmb*vmb)
{
	u8 txitlr = readb((void*)(0x0356 + vmb->addr));

	return txitlr;
}
inline void write_txitlr(struct vmb*vmb,u8 txitlr)
{
	writeb(txitlr,(void*)(0x0356 + vmb->addr));
}
inline u8 read_txithr(struct vmb*vmb)
{
	u8 txithr = readb((void*)(0x0358 + vmb->addr));

	return txithr;
}
inline void write_txithr(struct vmb*vmb,u8 txithr)
{
	writeb(txithr,(void*)(0x0358 + vmb->addr));
}
inline u16 read_txit(struct vmb*vmb)
{
	u8 txitlr = readb((void*)(0x0356 + vmb->addr));
	u8 txithr = 0x3F & readb((void*)(0x0358 + vmb->addr));

	return txithr<<8|txitlr;
}
inline void write_txit(struct vmb*vmb,u16 txit)
{
	write_txitlr(vmb,txit&0x00FF);
	write_txithr(vmb,txit>>8);				
}
inline u8 read_rxdtor(struct vmb*vmb)
{
	u8 rxdtor = readb((void*)(0x035A + vmb->addr));

	return rxdtor;
}
inline u16 read_mplr(struct vmb*vmb,u8 n)
{
	ASSERT(n < 2);

	u16 mplr = readw((void*)(0x0380+n+n + vmb->addr));

	return mplr;
}
inline void write_mplr(struct vmb*vmb,u8 n,u16 mplr)
{
	ASSERT(n < 2);

	writew(mplr,(void*)(0x0380+n+n + vmb->addr));
}
/*-------------------------------------------------------------------------*/
//			  IO UTILITY FUNCTIONS
/*-------------------------------------------------------------------------*/
static void set_max_plen(struct vmb*vmb,u8 td,struct urb*urb)
{
	ASSERT(td < 16);

	struct vmb_slot*slot = &vmb->slot[td];

	ASSERT(slot->allocated);
	ASSERT(0==slot->initialized);

	u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));

	// note that maxusbpktlen can be bigger than urb->transfer_buffer_length

	u16 mpl = vmb_maxpacket_mpl(maxusbpktlen);
	u8 range = td>>3;
	u16 max_plen = read_mplr(vmb,range);
	u8 shift = 0x0E & (td<<1);
	max_plen &= ~(3<<shift);
	max_plen |= mpl<<shift;
	write_mplr(vmb,range,max_plen);

	slot->max_plen = vmb_maxpacket_len(maxusbpktlen);
}
static int check_max_plen(struct vmb*vmb,u8 td,struct urb*urb)
{
	ASSERT(td < 16);

	struct vmb_slot*slot = &vmb->slot[td];

	ASSERT(slot->allocated);

	u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));

	return slot->max_plen == vmb_maxpacket_len(maxusbpktlen);
}
static void update_max_plen(struct vmb*vmb,u8 td,struct urb*urb)
{
	ASSERT(td < 16);

	struct vmb_slot*slot = &vmb->slot[td];

	ASSERT(slot->allocated);
	ASSERT(slot->initialized);
	ASSERT(check_urb_is_in_slot(slot,urb));

	u16 maxusbpktlen = usb_maxpacket(urb->dev,urb->pipe,usb_pipeout(urb->pipe));

	// note that maxusbpktlen can be bigger than urb->transfer_buffer_length

	u16 mpl = vmb_maxpacket_mpl(maxusbpktlen);
	u8 range = td>>3;
	u16 max_plen = read_mplr(vmb,range);
	u8 shift = 0x0E & (td<<1);
	max_plen &= ~(3<<shift);
	max_plen |= mpl<<shift;
	write_mplr(vmb,range,max_plen);

	slot->max_plen = vmb_maxpacket_len(maxusbpktlen);
}
static void clear_max_plen(struct vmb*vmb,u8 td)
{
	ASSERT(td < 16);

	u8 range = td>>3;
	u16 max_plen = read_mplr(vmb,range);
	u8 shift = 0x0E & (td<<1);
	max_plen &= ~(3<<shift);
	write_mplr(vmb,range,max_plen);
}
/*-------------------------------------------------------------------------*/
